package main

import (
	"context"
	"net/http"
	"os"
	"os/signal"
	"sync"
	"syscall"
	"time"

	"git.joco.dk/sng/fermentord/internal/api"
	"git.joco.dk/sng/fermentord/internal/controllers"
	"git.joco.dk/sng/fermentord/internal/dal"
	"git.joco.dk/sng/fermentord/internal/hw"
	"git.joco.dk/sng/fermentord/internal/metrics"
	"git.joco.dk/sng/fermentord/pkg/temperature"
	"github.com/fsnotify/fsnotify"
	"github.com/getsentry/sentry-go"
	"github.com/prometheus/client_golang/prometheus/promhttp"
	"github.com/rs/zerolog/log"
	"github.com/spf13/viper"
)

type configuration struct {
	Db struct {
		Dsn string `mapstructure:"dsn"`
	} `mapstructure:"db"`
	Temperature struct {
		WortSensor    string `mapstructure:"wort_sensor"`
		ChamberSensor string `mapstructure:"chamber_sensor"`
		AmbientSensor string `mapstructure:"ambient_sensor"`
	} `mapstructure:"temperature"`
	Hysteresis controllers.Config `mapstructure:"hysteresis"`
}

var (
	config     configuration
	ctx        context.Context
	wg         *sync.WaitGroup
	hysteresis *controllers.Hysteresis
	gpio       *hw.Gpio
)

func bool2ChamberState(state bool) dal.ChamberState {
	if state {
		return dal.ChamberStateCooling
	}
	return dal.ChamberStateIdle
}

func mainLoop(db *dal.DAL) {
	wg.Add(1)
	defer wg.Done()

	for {
		select {
		case <-ctx.Done():
			return

		case state := <-hysteresis.C:
			ctx := context.Background()

			if state {
				metrics.State.Set(1)
				//gpio.StartCooler()
			} else {
				metrics.State.Set(0)
				//gpio.StopCooler()
			}

			dbCtx, cancel := context.WithTimeout(ctx, 10*time.Second)
			tx, err := db.Begin()
			if err != nil {
				sentry.CaptureException(err)
				log.Error().Err(err).Msg("Failed to start DB transaction")
			} else {
				dal.SaveChamberState(dbCtx, tx, bool2ChamberState(state))
				err = tx.Commit()
				if err != nil {
					sentry.CaptureException(err)
					log.Error().Err(err).Msg("Failed to commit DB transaction")
				}
			}
			cancel()

		case reading := <-temperature.C:
			ctx := context.Background()

			switch reading.Sensor {
			case config.Temperature.AmbientSensor:
				hysteresis.UpdateAmbientTemperature(reading.Value)

			case config.Temperature.ChamberSensor:
				hysteresis.UpdateChamberTemperature(reading.Value)

			case config.Temperature.WortSensor:
				hysteresis.UpdateWortTemperature(reading.Value)
			}

			dbCtx, cancel := context.WithTimeout(ctx, 10*time.Second)
			tx, err := db.Begin()
			if err != nil {
				log.Error().Err(err).Msg("Failed to start DB transaction")
			} else {
				dal.SaveTemperatureReading(dbCtx, tx, reading.Sensor, reading.Value)
				err = tx.Commit()
				if err != nil {
					log.Error().Err(err).Msg("Failed to commit DB transaction")
				}
			}
			cancel()
		}
	}
}

func main() {
	// Sentry
	err := sentry.Init(sentry.ClientOptions{
		Dsn: "https://7278625538334140991ce433e0ad292f@sentry.joco.dk/24",
		//TracesSampleRate: 0.1,
	})
	if err != nil {
		log.Fatal().Err(err).Msg("Failed to initialize Sentry.")
	}
	defer sentry.Flush(10 * time.Second)

	// Configuration
	viper.AllowEmptyEnv(true)
	viper.AutomaticEnv()
	viper.SetDefault("db.dsn", "./fermentord.db")
	viper.AddConfigPath("/etc")
	viper.SetConfigName("fermentord")
	viper.SetConfigType("hcl")
	viper.ReadInConfig()
	viper.Unmarshal(&config)

	ctxb := context.Background()
	ctx, shutdown := context.WithCancel(ctxb)
	wg = &sync.WaitGroup{}

	temperature.Initialize(10 * time.Second)
	hysteresis = controllers.NewHysteresis("Chamber 1")

	reloadConfig := func() {
		temperature.SetSensors([]string{
			config.Temperature.AmbientSensor,
			config.Temperature.ChamberSensor,
			config.Temperature.WortSensor,
		})
		hysteresis.SetConfig(&config.Hysteresis)
	}

	viper.OnConfigChange(func(in fsnotify.Event) {
		reloadConfig()
	})
	viper.WatchConfig()
	reloadConfig()

	// Database
	log.Info().Str("dsn", config.Db.Dsn).Msg("Initializing DB")

	migrator, err := dal.NewMigrator(config.Db.Dsn)
	if err != nil {
		log.Fatal().Err(err).Msg("Error initializing DB migrations")
	}
	migrator.Up()
	serr, dberr := migrator.Close()
	if serr != nil {
		log.Fatal().Err(serr).Msg("DB migration source error")
	}
	if dberr != nil {
		log.Fatal().Err(dberr).Msg("DB migration database error")
	}

	var db *dal.DAL
	if db, err := dal.NewDAL(config.Db.Dsn); err != nil {
		log.Fatal().Err(err).Msg("Error connecting to DB")
	} else {
		db.Initialize()
		defer db.Close()
	}

	// GPIO
	log.Info().Msg("Initializing GPIO")
	gpio, err := hw.NewGpio()
	if err != nil {
		log.Fatal().Err(err).Send()
	}
	defer gpio.Close()

	// HTTP
	mux := http.NewServeMux()
	srv := &http.Server{
		Addr:    ":8000",
		Handler: mux,
	}

	api := api.NewAPI(db)

	// Metrics
	mux.Handle("/metrics", promhttp.Handler())
	mux.HandleFunc("/health", api.HealthCheck)
	mux.HandleFunc("/temperatures", api.GetTemperatures)
	mux.HandleFunc("/state-changes", api.GetStateChanges)

	// Main
	go temperature.Serve(ctx, wg)
	go mainLoop(db)
	go srv.ListenAndServe()

	done := make(chan os.Signal, 1)
	signal.Notify(done, os.Interrupt, syscall.SIGINT, syscall.SIGTERM)
	<-done

	go func() {
		ctx, cancel := context.WithTimeout(ctxb, 10*time.Second)
		defer cancel()
		err := srv.Shutdown(ctx)
		if err != nil {
			log.Error().Err(err).Msg("Error shutting down HTTP server.")
		}
	}()

	// Initiate graceful shutdown.
	shutdown()

	// Wait for graceful shutdown.
	wg.Wait()

	// TODO PID for heating
	// TODO Data export
	// TODO SQL migrations https://github.com/golang-migrate/migrate/tree/master/database/sqlite
}
